@@ -2,7 +2,7 @@
 
   vm_core.h - 
 
-  $Author: nobu $
+  $Author: yugui $
   created at: 04/01/01 19:41:38 JST
 
   Copyright (C) 2004-2007 Koichi Sasada
@@ -14,13 +14,10 @@
 
 #define RUBY_VM_THREAD_MODEL 2
 
-#include <setjmp.h>
-
 #include "ruby/ruby.h"
-#include "ruby/signal.h"
 #include "ruby/st.h"
-#include "ruby/node.h"
 
+#include "node.h"
 #include "debug.h"
 #include "vm_opts.h"
 #include "id.h"
@@ -33,18 +30,23 @@
 #error "unsupported thread type"
 #endif
 
+#include <setjmp.h>
 #include <signal.h>
 
 #ifndef NSIG
-# ifdef DJGPP
-#  define NSIG SIGMAX
-# else
-#  define NSIG (_SIGMAX + 1)      /* For QNX */
-# endif
+# define NSIG (_SIGMAX + 1)      /* For QNX */
 #endif
 
 #define RUBY_NSIG NSIG
 
+#ifdef HAVE_STDARG_PROTOTYPES
+#include <stdarg.h>
+#define va_init_list(a,b) va_start(a,b)
+#else
+#include <varargs.h>
+#define va_init_list(a,b) va_start(a)
+#endif
+
 /*****************/
 /* configuration */
 /*****************/
@@ -88,45 +90,7 @@
 #define UNLIKELY(x) (x)
 #endif /* __GNUC__ >= 3 */
 
-#define ISEQ_TYPE_TOP    INT2FIX(1)
-#define ISEQ_TYPE_METHOD INT2FIX(2)
-#define ISEQ_TYPE_BLOCK  INT2FIX(3)
-#define ISEQ_TYPE_CLASS  INT2FIX(4)
-#define ISEQ_TYPE_RESCUE INT2FIX(5)
-#define ISEQ_TYPE_ENSURE INT2FIX(6)
-#define ISEQ_TYPE_EVAL   INT2FIX(7)
-#define ISEQ_TYPE_DEFINED_GUARD INT2FIX(8)
-
-#define CATCH_TYPE_RESCUE INT2FIX(1)
-#define CATCH_TYPE_ENSURE INT2FIX(2)
-#define CATCH_TYPE_RETRY  INT2FIX(3)
-#define CATCH_TYPE_BREAK  INT2FIX(4)
-#define CATCH_TYPE_REDO   INT2FIX(5)
-#define CATCH_TYPE_NEXT   INT2FIX(6)
-
-struct iseq_insn_info_entry {
-    unsigned short position;
-    unsigned short line_no;
-    unsigned short sp;
-};
-
-struct iseq_catch_table_entry {
-    VALUE type;
-    VALUE iseq;
-    unsigned long start;
-    unsigned long end;
-    unsigned long cont;
-    unsigned long sp;
-};
-
-#define INITIAL_ISEQ_COMPILE_DATA_STORAGE_BUFF_SIZE (512)
-
-struct iseq_compile_data_storage {
-    struct iseq_compile_data_storage *next;
-    unsigned long pos;
-    unsigned long size;
-    char *buff;
-};
+typedef unsigned long rb_num_t;
 
 struct iseq_compile_data_ensure_node_stack;
 
@@ -139,31 +103,9 @@ typedef struct rb_compile_option_struct {
     int instructions_unification;
     int stack_caching;
     int trace_instruction;
+    int debug_level;
 } rb_compile_option_t;
 
-struct iseq_compile_data {
-    /* GC is needed */
-    VALUE err_info;
-    VALUE mark_ary;
-    VALUE catch_table_ary;	/* Array */
-
-    /* GC is not needed */
-    struct iseq_label_data *start_label;
-    struct iseq_label_data *end_label;
-    struct iseq_label_data *redo_label;
-    VALUE current_block;
-    VALUE loopval_popped;	/* used by NODE_BREAK */
-    VALUE ensure_node;
-    VALUE for_iseq;
-    struct iseq_compile_data_ensure_node_stack *ensure_node_stack;
-    int cached_const;
-    struct iseq_compile_data_storage *storage_head;
-    struct iseq_compile_data_storage *storage_current;
-    int last_line;
-    int flip_cnt;
-    const rb_compile_option_t *option;
-};
-
 #if 1
 #define GetCoreDataFromValue(obj, type, ptr) do { \
     ptr = (type*)DATA_PTR(obj); \
@@ -175,12 +117,6 @@ struct iseq_compile_data {
 #define GetISeqPtr(obj, ptr) \
   GetCoreDataFromValue(obj, rb_iseq_t, ptr)
 
-typedef struct rb_iseq_profile_struct {
-    VALUE count;
-    VALUE time_self;
-    VALUE time_cumu; /* cumulative */
-} rb_iseq_profile_t;
-
 struct rb_iseq_struct;
 
 struct rb_iseq_struct {
@@ -195,6 +131,8 @@ struct rb_iseq_struct {
     VALUE *iseq_encoded; /* encoded iseq */
     unsigned long iseq_size;
     VALUE mark_ary;	/* Array: includes operands which should be GC marked */
+    VALUE coverage;     /* coverage array */
+    unsigned short line_no;
 
     /* insn info, must be freed */
     struct iseq_insn_info_entry *insn_info_table;
@@ -254,6 +192,7 @@ struct rb_iseq_struct {
     /****************/
 
     VALUE self;
+    VALUE orig;			/* non-NULL if its data have origin */
 
     /* block inlining */
     /* 
@@ -269,12 +208,18 @@ struct rb_iseq_struct {
 
     /* misc */
     ID defined_method_id;	/* for define_method */
-    rb_iseq_profile_t profile;
 
     /* used at compile time */
     struct iseq_compile_data *compile_data;
 };
 
+enum ruby_special_exceptions {
+    ruby_error_reenter,
+    ruby_error_nomemory,
+    ruby_error_sysstack,
+    ruby_special_error_count
+};
+
 typedef struct rb_iseq_struct rb_iseq_t;
 
 #define GetVMPtr(obj, ptr) \
@@ -283,32 +228,47 @@ typedef struct rb_iseq_struct rb_iseq_t;
 typedef struct rb_vm_struct {
     VALUE self;
 
-    rb_thread_lock_t global_interpreter_lock;
+    rb_thread_lock_t global_vm_lock;
 
     struct rb_thread_struct *main_thread;
     struct rb_thread_struct *running_thread;
 
     st_table *living_threads;
     VALUE thgroup_default;
-    VALUE last_status; /* $? */
 
+    int running;
     int thread_abort_on_exception;
     unsigned long trace_flag;
+    volatile int sleeper;
 
     /* object management */
     VALUE mark_object_ary;
 
+    VALUE special_exceptions[ruby_special_error_count];
+
     /* load */
     VALUE top_self;
+    VALUE load_path;
     VALUE loaded_features;
     struct st_table *loading_table;
     
     /* signal */
-    rb_atomic_t signal_buff[RUBY_NSIG];
-    rb_atomic_t buffered_signal_size;
+    struct {
+	VALUE cmd;
+	int safe;
+    } trap_list[RUBY_NSIG];
 
     /* hook */
     rb_event_hook_t *event_hooks;
+
+    int src_encoding_index;
+
+    VALUE verbose, debug, progname;
+    VALUE coverages;
+
+#if defined(ENABLE_VM_OBJSPACE) && ENABLE_VM_OBJSPACE
+    struct rb_objspace *objspace;
+#endif
 } rb_vm_t;
 
 typedef struct {
@@ -324,11 +284,9 @@ typedef struct {
     VALUE proc;			/* cfp[9] / block[4] */
     ID method_id;               /* cfp[10] saved in special case */
     VALUE method_class;         /* cfp[11] saved in special case */
-    VALUE prof_time_self;       /* cfp[12] */
-    VALUE prof_time_chld;       /* cfp[13] */
 } rb_control_frame_t;
 
-typedef struct {
+typedef struct rb_block_struct {
     VALUE self;			/* share with method frame if it's only block */
     VALUE *lfp;			/* share with method frame if it's only block */
     VALUE *dfp;			/* share with method frame if it's only block */
@@ -343,10 +301,11 @@ enum rb_thread_status {
     THREAD_TO_KILL,
     THREAD_RUNNABLE,
     THREAD_STOPPED,
-    THREAD_KILLED,
+    THREAD_STOPPED_FOREVER,
+    THREAD_KILLED
 };
 
-typedef jmp_buf rb_jmpbuf_t;
+typedef RUBY_JMP_BUF rb_jmpbuf_t;
 
 struct rb_vm_tag {
     rb_jmpbuf_t buf;
@@ -362,9 +321,14 @@ struct rb_vm_trap_tag {
 #define RUBY_VM_VALUE_CACHE_SIZE 0x1000
 #define USE_VALUE_CACHE 0
 
-typedef struct rb_thread_struct rb_thread_t;
+struct rb_unblock_callback {
+    rb_unblock_function_t *func;
+    void *arg;
+};
 
-struct rb_thread_struct
+struct rb_mutex_struct;
+
+typedef struct rb_thread_struct
 {
     VALUE self;
     rb_vm_t *vm;
@@ -375,6 +339,7 @@ struct rb_thread_struct
     rb_control_frame_t *cfp;
     int safe_level;
     int raised_flag;
+    VALUE last_status; /* $? */
     
     /* passing state */
     int state;
@@ -396,8 +361,10 @@ struct rb_thread_struct
     rb_thread_id_t thread_id;
     enum rb_thread_status status;
     int priority;
+    int slice;
 
     native_thread_data_t native_thread_data;
+    void *blocking_region_buffer;
 
     VALUE thgroup;
     VALUE value;
@@ -407,14 +374,17 @@ struct rb_thread_struct
     int exec_signal;
 
     int interrupt_flag;
-    rb_unblock_function_t *unblock_function;
-    void *unblock_function_arg;
     rb_thread_lock_t interrupt_lock;
+    struct rb_unblock_callback unblock;
+    VALUE locking_mutex;
+    struct rb_mutex_struct *keeping_mutexes;
+    int transition_for_lock;
 
     struct rb_vm_tag *tag;
     struct rb_vm_trap_tag *trap_tag;
 
     int parse_in_eval;
+    int mild_compile_error;
 
     /* storage */
     st_table *local_storage;
@@ -458,17 +428,24 @@ struct rb_thread_struct
     /* misc */
     int method_missing_reason;
     int abort_on_exception;
-};
+} rb_thread_t;
 
 /* iseq.c */
 VALUE rb_iseq_new(NODE*, VALUE, VALUE, VALUE, VALUE);
-VALUE rb_iseq_new_with_bopt(NODE*, VALUE, VALUE, VALUE, VALUE, VALUE);
-VALUE rb_iseq_new_with_opt(NODE*, VALUE, VALUE, VALUE, VALUE, const rb_compile_option_t*);
+VALUE rb_iseq_new_top(NODE *node, VALUE name, VALUE filename, VALUE parent);
+VALUE rb_iseq_new_main(NODE *node, VALUE filename);
+VALUE rb_iseq_new_with_bopt(NODE*, VALUE, VALUE, VALUE, VALUE, VALUE, VALUE);
+VALUE rb_iseq_new_with_opt(NODE*, VALUE, VALUE, VALUE, VALUE, VALUE, const rb_compile_option_t*);
 VALUE rb_iseq_compile(VALUE src, VALUE file, VALUE line);
-VALUE ruby_iseq_disasm(VALUE self);
-VALUE ruby_iseq_disasm_insn(VALUE str, VALUE *iseqval, int pos, rb_iseq_t *iseq, VALUE child);
+VALUE rb_iseq_disasm(VALUE self);
+VALUE rb_iseq_disasm_insn(VALUE str, VALUE *iseqval, int pos, rb_iseq_t *iseq, VALUE child);
 const char *ruby_node_name(int node);
+int rb_iseq_first_lineno(rb_iseq_t *iseq);
 
+RUBY_EXTERN VALUE rb_cISeq;
+RUBY_EXTERN VALUE rb_cRubyVM;
+RUBY_EXTERN VALUE rb_cEnv;
+RUBY_EXTERN VALUE rb_mRubyVMFrozenCore;
 
 /* each thread has this size stack : 128KB */
 #define RUBY_VM_THREAD_STACK_SIZE (128 * 1024)
@@ -489,8 +466,6 @@ typedef struct {
     int safe_level;
     int is_from_method;
     int is_lambda;
-
-    NODE *special_cref_stack;
 } rb_proc_t;
 
 #define GetEnvPtr(obj, ptr) \
@@ -509,7 +484,6 @@ typedef struct {
 
 typedef struct {
     VALUE env;
-    NODE *cref_stack;
 } rb_binding_t;
 
 
@@ -523,6 +497,32 @@ typedef struct {
 #define VM_CALL_SUPER_BIT          (0x01 << 7)
 #define VM_CALL_SEND_BIT           (0x01 << 8)
 
+#define VM_SPECIAL_OBJECT_VMCORE   0x01
+#define VM_SPECIAL_OBJECT_CBASE    0x02
+
+#define VM_FRAME_MAGIC_METHOD 0x11
+#define VM_FRAME_MAGIC_BLOCK  0x21
+#define VM_FRAME_MAGIC_CLASS  0x31
+#define VM_FRAME_MAGIC_TOP    0x41
+#define VM_FRAME_MAGIC_FINISH 0x51
+#define VM_FRAME_MAGIC_CFUNC  0x61
+#define VM_FRAME_MAGIC_PROC   0x71
+#define VM_FRAME_MAGIC_IFUNC  0x81
+#define VM_FRAME_MAGIC_EVAL   0x91
+#define VM_FRAME_MAGIC_LAMBDA 0xa1
+#define VM_FRAME_MAGIC_MASK_BITS   8
+#define VM_FRAME_MAGIC_MASK   (~(~0<<VM_FRAME_MAGIC_MASK_BITS))
+
+#define VM_FRAME_TYPE(cfp) ((cfp)->flag & VM_FRAME_MAGIC_MASK)
+
+/* other frame flag */
+#define VM_FRAME_FLAG_PASSED 0x0100
+
+
+#define RUBYVM_CFUNC_FRAME_P(cfp) \
+  (VM_FRAME_TYPE(cfp) == VM_FRAME_MAGIC_CFUNC)
+
+
 /* inline (method|const) cache */
 #define NEW_INLINE_CACHE_ENTRY() NEW_WHILE(Qundef, 0, 0)
 #define ic_class  u1.value
@@ -546,8 +546,6 @@ typedef rb_control_frame_t *
 #define GC_GUARDED_PTR_REF(p) ((void *)(((VALUE)p) & ~0x03))
 #define GC_GUARDED_PTR_P(p)   (((VALUE)p) & 0x01)
 
-#define RUBY_VM_METHOD_NODE NODE_METHOD
-
 #define RUBY_VM_PREVIOUS_CONTROL_FRAME(cfp) (cfp+1)
 #define RUBY_VM_NEXT_CONTROL_FRAME(cfp) (cfp-1)
 #define RUBY_VM_END_CONTROL_FRAME(th) \
@@ -567,53 +565,37 @@ typedef rb_control_frame_t *
 #define RUBY_VM_GET_CFP_FROM_BLOCK_PTR(b) \
   ((rb_control_frame_t *)((VALUE *)(b) - 5))
 
-/* defined? */
-#define DEFINED_IVAR   INT2FIX(1)
-#define DEFINED_IVAR2  INT2FIX(2)
-#define DEFINED_GVAR   INT2FIX(3)
-#define DEFINED_CVAR   INT2FIX(4)
-#define DEFINED_CONST  INT2FIX(5)
-#define DEFINED_METHOD INT2FIX(6)
-#define DEFINED_YIELD  INT2FIX(7)
-#define DEFINED_REF    INT2FIX(8)
-#define DEFINED_ZSUPER INT2FIX(9)
-#define DEFINED_FUNC   INT2FIX(10)
-
 /* VM related object allocate functions */
-/* TODO: should be static functions */
 VALUE rb_thread_alloc(VALUE klass);
 VALUE rb_proc_alloc(VALUE klass);
 
 /* for debug */
-extern void vm_stack_dump_raw(rb_thread_t *, rb_control_frame_t *);
-#define SDR() vm_stack_dump_raw(GET_THREAD(), GET_THREAD()->cfp)
-#define SDR2(cfp) vm_stack_dump_raw(GET_THREAD(), (cfp))
+extern void rb_vmdebug_stack_dump_raw(rb_thread_t *, rb_control_frame_t *);
+#define SDR() rb_vmdebug_stack_dump_raw(GET_THREAD(), GET_THREAD()->cfp)
+#define SDR2(cfp) rb_vmdebug_stack_dump_raw(GET_THREAD(), (cfp))
 void rb_vm_bugreport(void);
 
 
 /* functions about thread/vm execution */
 
 VALUE rb_iseq_eval(VALUE iseqval);
+VALUE rb_iseq_eval_main(VALUE iseqval);
 void rb_enable_interrupt(void);
 void rb_disable_interrupt(void);
 int rb_thread_method_id_and_class(rb_thread_t *th, ID *idp, VALUE *klassp);
 
-VALUE vm_eval_body(rb_thread_t *th);
-VALUE vm_invoke_proc(rb_thread_t *th, rb_proc_t *proc, VALUE self,
-		     int argc, VALUE *argv, rb_block_t *blockptr);
-VALUE vm_make_proc(rb_thread_t *th, rb_control_frame_t *cfp, rb_block_t *block);
-VALUE vm_make_env_object(rb_thread_t *th, rb_control_frame_t *cfp);
-VALUE vm_backtrace(rb_thread_t *, int);
+VALUE rb_vm_invoke_proc(rb_thread_t *th, rb_proc_t *proc, VALUE self,
+			int argc, const VALUE *argv, rb_block_t *blockptr);
+VALUE rb_vm_make_proc(rb_thread_t *th, const rb_block_t *block, VALUE klass);
+VALUE rb_vm_make_env_object(rb_thread_t *th, rb_control_frame_t *cfp);
 
-VALUE vm_yield(rb_thread_t *th, int argc, VALUE *argv);
-VALUE vm_call0(rb_thread_t *th, VALUE klass, VALUE recv, VALUE id, ID oid,
-	       int argc, const VALUE *argv, NODE *body, int nosuper);
-
-int vm_get_sourceline(rb_control_frame_t *);
+void *rb_thread_call_with_gvl(void *(*func)(void *), void *data1);
+int ruby_thread_has_gvl_p(void);
+rb_control_frame_t *rb_vm_get_ruby_level_next_cfp(rb_thread_t *th, rb_control_frame_t *cfp);
 
 NOINLINE(void rb_gc_save_machine_context(rb_thread_t *));
 
-RUBY_EXTERN VALUE sysstack_error;
+#define sysstack_error GET_VM()->special_exceptions[ruby_error_sysstack]
 
 /* for thread */
 
@@ -623,7 +605,7 @@ extern rb_vm_t *ruby_current_vm;
 
 #define GET_VM() ruby_current_vm
 #define GET_THREAD() ruby_current_thread
-#define rb_thread_set_current_raw(th) (ruby_current_thread = th)
+#define rb_thread_set_current_raw(th) (void)(ruby_current_thread = (th))
 #define rb_thread_set_current(th) do { \
     rb_thread_set_current_raw(th); \
     th->vm->running_thread = th; \
@@ -635,13 +617,13 @@ extern rb_vm_t *ruby_current_vm;
 
 #define RUBY_VM_SET_INTERRUPT(th) ((th)->interrupt_flag |= 0x02)
 #define RUBY_VM_SET_TIMER_INTERRUPT(th) ((th)->interrupt_flag |= 0x01)
+#define RUBY_VM_SET_FINALIZER_INTERRUPT(th) ((th)->interrupt_flag |= 0x04)
 #define RUBY_VM_INTERRUPTED(th) ((th)->interrupt_flag & 0x02)
 
 void rb_thread_execute_interrupts(rb_thread_t *);
 
 #define RUBY_VM_CHECK_INTS_TH(th) do { \
-  if(th->interrupt_flag){ \
-    /* TODO: trap something event */ \
+  if (UNLIKELY(th->interrupt_flag)) { \
     rb_thread_execute_interrupts(th); \
   } \
 } while (0)
@@ -650,29 +632,14 @@ void rb_thread_execute_interrupts(rb_thread_t *);
   RUBY_VM_CHECK_INTS_TH(GET_THREAD())
 
 /* tracer */
-static void inline
-exec_event_hooks(rb_event_hook_t *hook, rb_event_flag_t flag, VALUE self, ID id, VALUE klass)
-{
-    while (hook) {
-	if (flag & hook->flag) {
-	    (*hook->func)(flag, hook->data, self, id, klass);
-	}
-	hook = hook->next;
-    }
-}
+void
+rb_threadptr_exec_event_hooks(rb_thread_t *th, rb_event_flag_t flag, VALUE self, ID id, VALUE klass);
 
 #define EXEC_EVENT_HOOK(th, flag, self, id, klass) do { \
     rb_event_flag_t wait_event__ = th->event_flags; \
     if (UNLIKELY(wait_event__)) { \
 	if (wait_event__ & (flag | RUBY_EVENT_VM)) { \
-	    VALUE self__ = (self), klass__ = (klass); \
-	    ID id__ = (id); \
-	    if (wait_event__ & flag) { \
-		exec_event_hooks(th->event_hooks, flag, self__, id__, klass__); \
-	    } \
-	    if (wait_event__ & RUBY_EVENT_VM) { \
-		exec_event_hooks(th->vm->event_hooks, flag, self__, id__, klass__); \
-	    } \
+	    rb_threadptr_exec_event_hooks(th, flag, self, id, klass); \
 	} \
     } \
 } while (0)
